#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/input.h>
#include <linux/kd.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include "names.h"

#define EVDEV          "/dev/input/event"
#define LIRCDEV        "/dev/lircd"
#define PROGNAME       "activylircd"
#define VERSION        "0.3"

#define PLAY_PAUSE_KEY 0x8f
#define MAX_VALUE      0xff

typedef struct client {
   int fd;
   struct client *next;
} client_t;

static client_t *clients = NULL;

typedef struct keycode_lircmap {
   int keycode;
   char * lircString;
} key2lirc;


static key2lirc keyTable[ MAX_VALUE + 1 ];

/* prototypes */
static void no_linger(int sock);
static int setup_key_table(char *mapFile);
static int setup_lirc_socket(void);
static int setup_event_dev(void);
static void process_key(int efd);
static void process_new_client(int sockfd);
static int is_a_console(int fd);
static int open_a_console(const char *fnam);
static int get_console(void);


static void no_linger(int sock)
{
   static struct linger  linger = {0, 0};
   int lsize  = sizeof(struct linger);
   setsockopt(sock, SOL_SOCKET, SO_LINGER, (void *)&linger, lsize);
}


static int setup_key_table(char * mapFile)
{
   int i, conFd;
   FILE *fd;
   char line[256];

   for( i = 1; i <= MAX_VALUE; i ++ ) {
      keyTable[i].keycode = 0;
      keyTable[i].lircString = NULL;
   }
   keyTable[0].lircString = "pause";

   fd = fopen(mapFile, "r");
   if( fd == NULL ) {
      fprintf(stderr, "Cannot open <%s>\n", mapFile);
      return 1;
   }
   while( !feof(fd) ) {
      fgets(line, sizeof(line)-1, fd);
      if(line[0] != '#' && line[0] != ';') {
         char *p = strchr( line, '=' );
         if(p && (p > line) && (*(p+1) != '\0')) {
            int key = 0;
            *p = '\0';
            key = atoi(line);            
            p++;
            *(p+strlen(p)-1)='\0';
            keyTable[key].lircString = strdup(p);
            printf( "Key: x%X - Lirc: <%s>\n", key, keyTable[key].lircString );
         }
      }
   }
   
   /* Setting keycodes 
   conFd = get_console();
   if( conFd >= 0 ) {
      struct kbkeycode a;

      for( i = 0; i <= MAX_VALUE; i++ ) {
         if( keyTable[i].lircString != NULL ) {
            a.keycode = 120;
            a.scancode = i;
            if( ioctl(conFd,KDSETKEYCODE,&a) )
               fprintf(stderr,"failed to set scancode %x to keycode %d\n",
                       a.scancode, a.keycode);
         } 
      }
      close(conFd);
   }
   */
   return 0;
}

static int setup_lirc_socket()
{
   int sockfd;
   struct sockaddr_un serv_addr;

   /* create socket */
   sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
   if( sockfd == -1 ) {
      fprintf(stderr, "%s: could not create socket\n", PROGNAME);
      perror(PROGNAME);
      return -1;
   }
   unlink(LIRCDEV);

   serv_addr.sun_family = AF_UNIX;
   strcpy(serv_addr.sun_path, LIRCDEV);
   if( bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) == -1 ) {
      fprintf(stderr, "%s: could not assign address to socket\n",PROGNAME);
      return -1;
   }

   chmod(LIRCDEV, 0666);
   listen(sockfd, 3);
   no_linger(sockfd);

   return sockfd;
}

static int setup_event_dev()
{
   int rfd = -1;
   int i;
   char name[256];
   char info[128];

   for( i = 0; i < 16; i++ ) {
      sprintf( name, "%s%d", EVDEV, i );
      if( (rfd = open (name, O_RDONLY)) >= 0 ) {
         ioctl (rfd, EVIOCGNAME (sizeof (info)), info);
         if( strstr( info, "keyboard" ) ) {
            printf("Using: %s (%s)\n", name, info);
            break;
         }
         printf("Not using: %s (%s)\n", name, info);
         close( rfd );
         rfd = -1;
      }
      else
         fprintf(stderr,"Can't open input event device %s\n", name);
   }
   return rfd;
}

static void process_key(int efd)
{
   static int lastKey = 0;
   struct input_event event[64];
   char message[1000];
   int len, idx;
   int broken = 0;
   client_t *client, *prev, *next;

   len = read(efd, event, sizeof(struct input_event) * 64);

   if( len < (int) sizeof(struct input_event) ) {
      fprintf(stderr, "Error processing input event %s\n", strerror(errno));
      return;
   }

//    for (i = 0; i < (int) (len / sizeof(struct input_event)); i++)
//    {
//       printf("Event: time %ld.%06ld, type %d, code %d, value %d\n",
//              event[i].time.tv_sec, event[i].time.tv_usec, event[i].type,
//              event[i].code, event[i].value);
//
//    }
   if( event[1].value < 1 || event[0].value > MAX_VALUE )
      return;
//
   printf ("%02x 0x%04x %04x\n", event[0].type, event[0].value, event[0].code);
   idx = event[0].value;
   if( keyTable[idx].lircString != NULL ) {
      if( lastKey == PLAY_PAUSE_KEY && idx == PLAY_PAUSE_KEY )
         lastKey = 0;
      else
         lastKey = idx;

      len = snprintf(message, sizeof message, "%016x %02x %s ACTIVY\n", event[0].value, 0, keyTable[lastKey].lircString);
   }
   else {
      if( KEY_NAME[event[1].code] )
         len = snprintf(message, sizeof message, "%016x %02x %s ACTIVY\n", event[1].code, 0, KEY_NAME[event[1].code]);
      else
         len = snprintf(message, sizeof message, "%016x %02x KEY_CODE_%d ACTIVY\n", event[1].code, 0, event[1].code);
   }

   for( client = clients; client; client = client->next ) {
      if( write(client->fd, message, len) != len ) {
         close(client->fd);
         client->fd = -1;
         broken = 1;
      }
   }

   if( broken ) {
      for( prev = NULL, client = clients; client; client = next ) {
         next = client->next;
         if( client->fd < 0 ) {
            if( prev )
               prev->next = client->next;
            else
               clients = client->next;
            free(client);
         }
         else {
            prev = client;
         }
      }
   }
   return;
}

static void process_new_client(int sockfd) 
{
   client_t *newclient = calloc(sizeof *newclient, 1);

   newclient->fd = accept(sockfd, NULL, NULL);

   if( newclient->fd < 0 ) {
      free(newclient);
      if( errno == ECONNABORTED || errno == EINTR )
         return;
      fprintf(stderr, "Error during accept(): %s\n", strerror(errno));
      return;
   }

   int flags = fcntl(newclient->fd, F_GETFL);
   fcntl(newclient->fd, F_SETFL, flags | O_NONBLOCK);
   no_linger(newclient->fd);

   newclient->next = clients;
   clients = newclient;
}


int main( int argc, char **argv )
{
   fd_set permset, fdset;
   int max_fd, event_fd, lirc_fd;

   printf( "\n%s Version %s\n", PROGNAME, VERSION );
   event_fd = setup_event_dev();
   if( event_fd < 0 ) {
      fprintf(stderr,"No input device found\n" );
      return 1;
   }

   if( setup_key_table( argv[1] )) {
      fprintf(stderr,"Error reading <argv[1]>\n" );
      return 1;
   }

   //GetKey(rfd, message);
   lirc_fd = setup_lirc_socket();
   if( lirc_fd < 0 ) {
      fprintf(stderr,"Can't init lirc socket.\n");
      return -1;
   }

   signal(SIGPIPE, SIG_IGN);

   FD_ZERO(&permset);
   FD_SET(event_fd, &permset);
   FD_SET(lirc_fd, &permset);
   if( event_fd > lirc_fd )
      max_fd = event_fd;
   else
      max_fd = lirc_fd;

   max_fd++;

   while( 1 ) {
      fdset = permset;

      if( select(max_fd, &fdset, NULL, NULL, NULL) < 0 ) {
         if( errno == EINTR )
            continue;
         fprintf(stderr, "Error during select(): %s\n", strerror(errno));
         return -1;
      }

      if( FD_ISSET(event_fd, &fdset) )
         process_key(event_fd);

      if( FD_ISSET(lirc_fd, &fdset) )
         process_new_client(lirc_fd);
   }

   unlink(LIRCDEV);
   return 0;
}


static int is_a_console(int fd) 
{
   char arg;

   arg = 0;
   return(ioctl(fd, KDGKBTYPE, &arg) == 0
          && ((arg == KB_101) || (arg == KB_84)));
}

static int open_a_console(const char *fnam) 
{
   int fd;

   /*
    * For ioctl purposes we only need some fd and permissions
    * do not matter. But setfont:activatemap() does a write.
    */
   fd = open(fnam, O_RDWR);
   if( fd < 0 && errno == EACCES )
      fd = open(fnam, O_WRONLY);
   if( fd < 0 && errno == EACCES )
      fd = open(fnam, O_RDONLY);
   if( fd < 0 )
      return -1;
   if( !is_a_console(fd) ) {
      close(fd);
      return -1;
   }
   return fd;
}

static int get_console( void )
{
   int fd;

   fd = open_a_console("/dev/tty");
   if( fd >= 0 )
      return fd;

   fd = open_a_console("/dev/tty0");
   if( fd >= 0 )
      return fd;

   fd = open_a_console("/dev/vc/0");
   if( fd >= 0 )
      return fd;

   fd = open_a_console("/dev/console");
   if( fd >= 0 )
      return fd;

   for( fd = 0; fd < 3; fd++ )
      if( is_a_console(fd) )
         return fd;

   fprintf(stderr, "Couldnt get a file descriptor referring to the console\n");
   return -1;
}
